from pathlib import Path

home = Path.home()
files = {
    'history': f'{home}/span/history',
    'instances': f'{home}/span/instances',
    'key': f'{home}/span/key',
    'account': f'{home}/span/account'
}



from getpass import getpass
def tutorial():
    print("""\033[33mYou are now in the tutorial, in which the data you input will not be saved.\033[0m
[Press Ctrl+C to exit the tutorial]
Span is an easy-to-use tool for creating and managing IKEv2 VPN servers in your Vultr account.
\033[1;34m[Task 1]\033[0m Before using Span, make sure you have a Vultr account \033[1mwith a valid payment method\033[0m.
Then go to https://my.vultr.com/settings/#settingsapi to get your API key.
In a Span shell, use this command to set API key:
    key
In this case, \033[7mLVGXWIIO6US43SR41U4KG3P552MHIDC5LURT\033[0m is your API key. Now try out the new command you have learnt!""")
    while 1:
        if (cmd := input('(tutorial) span\033[32m$\033[0m ')) == 'key':
            if getpass('API Key: ') == 'LVGXWIIO6US43SR41U4KG3P552MHIDC5LURT':
                print('\033[32mWell done!\033[0m')
                break
        elif cmd == '':
            print('\033[33mTask 1 skipped.\033[0m')
            break
        print('\033[33mPlease read task 1 carefully and try again.\033[0m')
    print("""\033[1;34m[Task 2]\033[0m Now set your beloved username for VPN connection, using this command:
    user <username> (e.g. `user john`)
In this case, \033[7madmin\033[0m will be your username.""")
    while 1:
        if (cmd := input('(tutorial) span\033[32m$\033[0m ')) == 'user admin':
            print('\033[32mWell done!\033[0m')
            break
        elif cmd == '':
            print('\033[33mTask 2 skipped.\033[0m')
            break
        print('\033[33mPlease read task 2 carefully and try again.\033[0m')
    print("""\033[1;34m[Task 3]\033[0m Now set a strong password for VPN connection, using this command:
    pwd
In this case, \033[7m3QJ2Mj2yCH98P8Lz\033[0m will be your password.""")
    while 1:
        if (cmd := input('(tutorial) span\033[32m$\033[0m ')) == 'pwd':
            while 1:
                tmp = getpass('Password: ')
                if tmp == '':
                    print('Error: Password cannot be empty.')
                    continue
                if tmp == getpass('Confirm Password: '):
                    break
                print('Error: Passwords do not match.')
            if tmp == '3QJ2Mj2yCH98P8Lz':
                print('\033[32mWell done!\033[0m')
                break
        elif cmd == '':
            print('\033[33mTask 3 skipped.\033[0m')
            break
        print('\033[33mPlease read task 3 carefully and try again.\033[0m')
    print("""\033[1;34m[Task 4]\033[0m Above are steps to initialize Span. Once you have finished those steps, there is no need to do it again, even if you reinstall Span.
It is time to introduce one of the most important commands. To create \033[1mand\033[0m setup a new instance, use this command:
    create <name> <region> <plan> (e.g. `create myinstance sjc vhp-1c-1gb-amd`)
You may be curious about where to find the region ID and plan ID. Use `regions` for a list of all regions and `plans` for a list of all plans (not included in the tutorial).
Supposing you would like to create a instance named \033[7mtest\033[0m in Singapore (whose ID is \033[7msgp\033[0m) with the plan \033[7mvhp-1c-1gb-intel\033[0m, try out the new command you have learnt.""")
    while 1:
        if (cmd := input('(tutorial) span\033[32m$\033[0m ')) == 'create test sgp vhp-1c-1gb-intel':
            print('\033[32mWell done!\033[0m')
            break
        elif cmd == '':
            print('\033[33mTask 4 skipped.\033[0m')
            break
        print('\033[33mPlease read task 4 carefully and try again.\033[0m')
    print("""\033[1;34m[Task 5]\033[0m Now you have a new instance named \033[7mtest\033[0m. However, in a real-world scenario, the creation process can be interrupted unexpectedly (by user, by network outage, etc.). In this case, you can use `setup <name>` to setup an existing instance manually (not included in the tutorial).
For a list of all your instances, use `ls`; for information about a specific instance, use `info <name>` (e.g. `info myinstance`). Neither is included in the tutorial.
The way to connect VPNs created with Span varies depending on the operating system of your device, but anyway, you definitely need the IP of your instance (can be found via `ls` or `info <name>`) as the VPN's server address, and the username and password you set for authentication. Plus, you need to have Span VPN CA installed on your device, which you can export with the command `export-ca <target_directory>` (e.g. `export-ca /Users/john/Desktop`) (not included in the tutorial). Google for more tips on VPN connection.
Now, let's try deleting the instance we just created, using this command:
    del <name> (e.g. `del myinstance`)""")
    while 1:
        if (cmd := input('(tutorial) span\033[32m$\033[0m ')) == 'del test':
            print('\033[32mWell done!\033[0m')
            break
        elif cmd == '':
            print('\033[33mTask 5 skipped.\033[0m')
            break
        print('\033[33mPlease read task 5 carefully and try again.\033[0m')
    print("""\033[1;34m[Finished]\033[0m Congratulations! You have successfully completed the tutorial.
Type `help` whenever you forget the usage of a command, and `tutorial` to restart the interactive tutorial.
To exit Span gracefully, use `exit` or press Ctrl+D. That's all!
\033[33mYou have quit the tutorial.\033[0m""")



from json import load
from os import remove
from paramiko import AutoAddPolicy, SSHClient
from paramiko.ssh_exception import NoValidConnectionsError
from scp import SCPClient
from socket import timeout
from .properties import files

def setup(instance):
    from random import choice
    with open(files['account'], 'r') as f:
        account = load(f)
    user = account['username']
    passwd = account['password']
    dict_ = list('qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM1234567890')
    psk = ''
    for i in range(64):
        psk += choice(dict_)
    with open('ipsec.secrets.sample') as f:
        with open('span.data/ipsec.secrets', 'w') as g:
            g.write(f.read().replace('$psk', psk).replace('$passwd', passwd).replace('$user', user))
    ssh = SSHClient()
    ssh.set_missing_host_key_policy(AutoAddPolicy())
    host = instance['ip']
    user = 'root'
    port = 22
    print('It may take a few minutes for the instance to get ready for further operations...')
    while 1:
        try:
            ssh.connect(host, port, user, instance['pwd'], timeout=3)
            break
        except (timeout, NoValidConnectionsError):
            pass
    print('Executing setup script...')
    scp = SCPClient(ssh.get_transport())
    scp.put('span.data', '.', True)
    remove('span.data/ipsec.secrets')
    ssh.exec_command('cd span.data ; chmod 755 -R . ; ./setup > setup.log 2>&1')
    ssh.close()



from getpass import getpass
from json import dump, load
from os import chdir, makedirs
from os.path import dirname, exists
from requests import get, post, delete
from time import sleep
import readline
from rich.console import Console
from rich.table import Column, Table
from .properties import files
from .setup import setup
from .tutorial import tutorial

chdir(dirname(__file__))
default = {'history': '', 'instances': '[]', 'key': '', 'account': '{}'}
for file in default:
    if not exists(files[file]):
        makedirs(dirname(files[file]), exist_ok=True)
        with open(files[file], 'w') as f:
            f.write(default[file])
readline.read_history_file(files['history'])
with open(files['key'], 'r') as f:
    api_key = f.read()


def cli():
    pass


def set_api_key():
    global api_key
    while 1:
        tmp = getpass('API Key: ')
        test = get('https://api.vultr.com/v2/account', headers={'Authorization': 'Bearer %s' % tmp})
        if test.status_code == 200:
            api_key = tmp
            break
        print('Error: %s' % test.json()['error'])
    with open(files['key'], 'w') as f:
        f.write(api_key)


def set_username(username):
    with open(files['account'], 'r') as f:
        original = load(f)
    original.update({'username': username})
    with open(files['account'], 'w') as f:
        dump(original, f)


def set_password():
    while 1:
        tmp = getpass('Password: ')
        if tmp == '':
            print('Error: Password cannot be empty.')
            continue
        if tmp == getpass('Confirm Password: '):
            break
        print('Error: Passwords do not match.')
    with open(files['account'], 'r') as f:
        original = load(f)
    original.update({'password': tmp})
    with open(files['account'], 'w') as f:
        dump(original, f)


def regions():
    lst = []
    cursor = ''
    params = {'per_page': 500}
    while 1:
        if cursor != '':
            params['cursor'] = cursor
        page = get('https://api.vultr.com/v2/regions', params).json()
        for region in page['regions']:
            lst.append(list(region.values())[:-1])
        if page['meta']['links']['next'] == '':
            break
        cursor = page['meta']['links']['next']
    return lst


def plans():
    lst = []
    cursor = ''
    params = {'per_page': 500, 'type': 'vhp'}
    region_ids = {i[0] for i in regions}
    while 1:
        if cursor != '':
            params['cursor'] = cursor
        page = get('https://api.vultr.com/v2/plans', params).json()
        for plan in page['plans']:
            lst.append([plan['id'], str(plan['bandwidth']), str(plan['monthly_cost']), ', '.join(region_ids - set(plan['locations']))])
        if page['meta']['links']['next'] == '':
            break
        cursor = page['meta']['links']['next']
    return lst


def create_instance(name, region, plan):
    with open(files['instances'], 'r') as f:
        for instance in load(f):
            if instance['n'] == name:
                print('Error: Instance with name "%s" already exists.' % name)
                return
    data = {'region': region, 'plan': plan, 'os_id': 381, 'label': 'Span Instance'}
    instance = post('https://api.vultr.com/v2/instances', json=data, headers={'Authorization': 'Bearer %s' % api_key})
    if instance.status_code != 202:
        print('Error: %s' % instance.json()['error'])
        return
    instance = instance.json()['instance']
    password = instance['default_password']
    for plan in plans:
        if plan[0] == instance['plan']:
            cost = plan[2]
            break
    while 1:
        instance = get('https://api.vultr.com/v2/instances/%s' % instance['id'], headers={'Authorization': 'Bearer %s' % api_key}).json()['instance']
        if instance['status'] == 'active':
            break
        sleep(1)
    instance = {'id': instance['id'], 'n': name, 'ip': instance['main_ip'], 'r': instance['region'], 'pwd': password, 'd': instance['date_created'], 'b': str(instance['allowed_bandwidth']), 'c': str(cost)}
    with open(files['instances'], 'r') as f:
        original = load(f)
    original.append(instance)
    with open(files['instances'], 'w') as f:
        dump(original, f)
    return instance


def list_instances():
    with open(files['instances'], 'r') as f:
        instances = load(f)
    if instances == []:
        print("You haven't created any instances yet.")
        return
    table = Table('Name', Column('IP', no_wrap=True), 'Region', 'Date Created', 'BW/GB·mo⁻¹', 'Cost/$·mo⁻¹', highlight=True)
    for instance in instances:
        table.add_row(instance['n'], instance['ip'], instance['r'], instance['d'], instance['b'], instance['c'])
    console.print(table)


def get_info(name, output):
    with open(files['instances'], 'r') as f:
        instances = load(f)
    for instance in instances:
        if instance['n'] == name:
            if output:
                table = Table('Name', Column('IP', no_wrap=True), 'Region', 'Date Created', 'BW/GB·mo⁻¹', 'Cost/$·mo⁻¹', highlight=True)
                table.add_row(instance['n'], instance['ip'], instance['r'], instance['d'], instance['b'], instance['c'])
                console.print(table)
            return instance
    print('Error: Instance with name "%s" does not exist.' % name)


def delete_instance(name):
    id = ''
    with open(files['instances'], 'r') as f:
        instances = load(f)
    for instance in instances:
        if instance['n'] == name:
            id = instance['id']
            break
    if id == '':
        print('Error: Instance with name "%s" does not exist.' % name)
        return
    rq = delete('https://api.vultr.com/v2/instances/%s' % id, headers={'Authorization': 'Bearer %s' % api_key})
    if rq.status_code != 204:
        print('Error: %s' % rq.json()['error'])
    else:
        instances.remove(instance)
        with open(files['instances'], 'w') as f:
            dump(instances, f)


def get_regions():
    global regions, regions_df
    if type(regions) != list:
        print('Fetching regions...')
        regions = regions()
        regions_df = Table(Column('ID', no_wrap=True), 'City', 'Country', 'Continent')
        for region in regions:
            regions_df.add_row(*region)

def get_plans():
    global plans, plans_df
    if type(plans) != list:
        print('Fetching plans...')
        plans = plans()
        plans_df = Table(Column('ID', no_wrap=True), Column('Bandwidth/GB·mo⁻¹', justify='right'), Column('Cost/$·mo⁻¹', justify='right'), 'Excluded Regions', highlight=True)
        for plan in plans:
            plans_df.add_row(*plan)


console = Console()
while 1:
    try:
        cmd = input('span\033[32m$\033[0m ').split()
        readline.write_history_file(files['history'])
        if cmd == []:
            continue
        cmd[0] = cmd[0].lower()
        if cmd[0] == 'regions':
            get_regions()
            console.print(regions_df)
        elif cmd[0] == 'plans':
            get_regions()
            get_plans()
            console.print(plans_df)
        elif cmd[0] == 'create':
            if api_key == '':
                set_api_key()
            if len(cmd) != 4:
                print('Usage: create <name> <region> <plan>')
            else:
                get_regions()
                get_plans()
                print('Creating instance...')
                instance = create_instance(*cmd[1:])
                if instance != None:
                    print('Instance created. IPv4 address: %s' % instance['ip'])
                    setup(instance)
        elif cmd[0] == 'setup':
            if len(cmd) != 2:
                print('Usage: setup <name>')
            else:
                instance = get_info(cmd[1], False)
                if instance != None:
                    setup(instance)
        elif cmd[0] == 'key':
            set_api_key()
        elif cmd[0] == 'user':
            if len(cmd) != 2:
                print('Usage: user <username>')
            else:
                set_username(cmd[1])
        elif cmd[0] == 'pwd':
            set_password()
        elif cmd[0] == 'ls':
            list_instances()
        elif cmd[0] == 'info':
            if len(cmd) != 2:
                print('Usage: info <name>')
            else:
                get_info(cmd[1], True)
        elif cmd[0] == 'del':
            if len(cmd) != 2:
                print('Usage: del <name>')
            else:
                delete_instance(cmd[1])
        elif cmd[0] == 'export-ca':
            if len(cmd) != 2:
                print('Usage: export-ca <target_directory>')
            else:
                with open('span.data/ca.cert.pem', 'rb') as s:
                    with open(cmd[1] + '/span-vpn-ca.pem', 'wb') as t:
                        t.write(s.read())
        elif cmd[0] == 'tutorial':
            tutorial()
        elif cmd[0] == 'help':
            print('''Usage:
    key                               Set API key.
    user <username>                   Set username for VPN connection.
    pwd                               Set password for VPN connection.
    regions                           List all regions.
    plans                             List all plans.
    create <name> <region> <plan>     Create and setup a new instance.
    setup <name>                      Setup an existing instance.
    ls                                List all instances.
    info <name>                       Get information about an instance.
    del <name>                        Delete an instance.
    export-ca <target_directory>      Export Span VPN CA.
    exit                              Exit Span.
    tutorial                          Enter interactive tutorial.
    help                              Show this message.''')
        elif cmd[0] == 'exit':
            break
        else:
            print('Error: Unknown command "%s"' % cmd[0])
    except EOFError:
        print()
        break
    except KeyboardInterrupt:
        print()
    except Exception as e:
        print('\033[31mError:\033[0m', e)
